#! /usr/bin/env python3
import rospy
import math
import numpy as np
from sensor_msgs.msg import LaserScan  # /scan topic (range finder)
from nav_msgs.msg import Odometry, OccupancyGrid
from geometry_msgs.msg import Pose,PoseArray
from copy import copy
from tf.transformations import euler_from_quaternion, quaternion_from_euler
import random

# Config parameters
dt = 0.01 # time between measurements

error = 0.01
toggleNoise = 1.0 # Change this value to 0 to remove noise

# You can change any of the following parameters
map_resolution = 0.02
map_x = 10
map_y = 10

grid_dim_x = ***
grid_dim_y = ***

N = *** # Number of particles

# Utils --- Do not change
# ----------------------------------------------------------------------------------------------------------
def get_rotation (msg):
    global roll, pitch, yaw
    orientation_q = msg.pose.pose.orientation
    orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
    (roll, pitch, yaw) = euler_from_quaternion (orientation_list)
    return yaw

def get_quaternion_from_euler(roll, pitch, yaw):
  qx = np.sin(roll/2) * np.cos(pitch/2) * np.cos(yaw/2) - np.cos(roll/2) * np.sin(pitch/2) * np.sin(yaw/2)
  qy = np.cos(roll/2) * np.sin(pitch/2) * np.cos(yaw/2) + np.sin(roll/2) * np.cos(pitch/2) * np.sin(yaw/2)
  qz = np.cos(roll/2) * np.cos(pitch/2) * np.sin(yaw/2) - np.sin(roll/2) * np.sin(pitch/2) * np.cos(yaw/2)
  qw = np.cos(roll/2) * np.cos(pitch/2) * np.cos(yaw/2) + np.sin(roll/2) * np.sin(pitch/2) * np.sin(yaw/2)

  return [qx, qy, qz, qw]


# Give the previous measurements and it will count displacement
def get_measurement(x,y,yaw):
    curr_pos = rospy.wait_for_message("/odom",Odometry,timeout=None)
    curr_yaw = get_rotation(curr_pos)
    dis_x = curr_pos.pose.pose.position.x-x + random.uniform(-error,error)*toggleNoise
    dis_y = curr_pos.pose.pose.position.y-y + random.uniform(-error,error)*toggleNoise
    d_yaw = curr_yaw-yaw + random.uniform(-error,error)*toggleNoise

    return dis_x, dis_y, d_yaw

def normalize_angle(angle):
    if angle < 0:
        return angle+np.ceil(-angle/(2*np.pi))*2*np.pi
    elif angle > 2*np.pi:
        return angle-(np.ceil(angle/(2*np.pi))-1)*2*np.pi
    return angle 
# ----------------------------------------------------------------------------------------------------------


def initialize_map_msg():
    map = OccupancyGrid()
    map.header.frame_id = "odom"
    map.info.resolution = map_resolution
    map.info.width  = ***
    map.info.height = ***
    map.info.origin.position.x = 0
    map.info.origin.position.y = 0
    map.info.origin.position.z = 0
    map.info.origin.orientation.x = 0
    map.info.origin.orientation.y = 0
    map.info.origin.orientation.z = 0    
    map.info.origin.orientation.w = 1
    
    # Add initial "Unknown values" to every cell:
    ***
    return map


def initialize_particles():
    ***


# Create a particles ROS-msg that will get published
def store_particles_msg(particles):
    ***
    return particles_msg


# Finds the particle with the heighest weight and store its map to the map msg
def store_map_msg(particles,map_msg):
    ***
    return map_msg


# You can use a class to represent each particle. 
class particle:

    def __init__(self):
        self.x   = 
        self.y   = 
        self.yaw = 
        
        self.weight = 

        self.map  = 



# Update the position of each particle given a certain measurement. 
def update_particles(particles,dx,dy,dyaw):
    ***
    return particles


# Updates the weights (and normalizes them)
def update_weights(particles,scan):
    ***
    return particles


# Update the occupancy grid of each particle
def update_OccGrid(particles,scan):
    ***
    return particles


# Perform low variance resampling. You can use copy() function 
def LowVarResample(particles):
    ***
    return new_particles


if __name__== "__main__":

    # Initialize ROS node
    rospy.init_node("FastSlam",anonymous=True)
    
    # Define ROS rate
    rate = rospy.Rate(1/dt)

    # Initialize particles (class)
    particles = initialize_particles()  

    # Initialize map msg 
    map_msg = initialize_map_msg()

    # Define publishers
    particles_pub = rospy.Publisher("/particles", PoseArray, queue_size=1)
    map_pub       = rospy.Publisher("/map", OccupancyGrid, queue_size=1)
    
    # Do not motify
    # -----------------------------------------------------------------------
    previous_position =  rospy.wait_for_message("/odom",Odometry,timeout=None)
    prev_x = previous_position.pose.pose.position.x
    prev_y = previous_position.pose.pose.position.y
    prev_yaw = get_rotation(previous_position)
    while not rospy.is_shutdown():
        # Measurements.
        dx,dy,dyaw   = get_measurement(prev_x,prev_y,prev_yaw)
        prev_x       = prev_x + dx
        prev_y       = prev_y + dy
        prev_yaw     = prev_yaw + dyaw
        scan =  rospy.wait_for_message("/scan",LaserScan,timeout=None)
    # -----------------------------------------------------------------------
        '''
        Your control input -> dx, dy, dyaw
        Your measurements  -> scan
        '''

        #Update particles positions
        particles = update_particles(copy(particles),dx,dy,dyaw)

        # Update weights
        particles = update_weights(copy(particles),scan)

        # Update map
        particles = update_OccGrid(copy(particles),scan)
        
        # Low-variance resampling
        particles = LowVarResample(copy(particles))

        # Prepare the particles_msg for it to be published
        particles_msg = store_particles_msg(particles)
        particles_pub.publish(particles_msg)

        # Publish the map of the particle with the highest weight
        map_msg = store_map_msg(particles,map_msg)
        map_pub.publish(map_msg)
        

        rate.sleep()  
